<?php
//Default config. nanoNodeName, nanoNodeRPCIP and nanoNodeAccount will be replaced by [ ./nl_run.py.py create ] command
$currency = 'nano';
$themeChoice = 'dark';
$nanoNodeName = 'nl_genesis'; #needs to be line 5
$nanoNodeRPCIP   = 'nl_genesis'; #needs to be line 6 
$nanoNodeRPCPort = '17076';
$nanoNodeAccount = 'nano_1fzwxb8tkmrp8o66xz7tcx65rm57bxdmpitw39ecomiwpjh89zxj33juzt6p'; #needs to be line 8
$nanoDonationAccount = 'nano_3msc38fyn67pgio16dj586pdrceahtn75qgnx7fy19wscixrc8dbb3abhbw6'; #https://mynano.ninja/account/gr0vity
$socials['github'] = 'https://github.com/gr0vity-dev/nano-local/';
$socials['reddit'] = 'https://reddit.com/u/gr0vity';
