export const environment = {
	production: false,
	network: 'test',
	wsUrl: 'ws://localhost:47000',
	rpcUrl: 'http://localhost:45000',
	principalsUrl: 'http://localhost:42002/json/monitors.json',
	explorerUrl: 'http://localhost:42000',
	repInfoUrl: 'http://localhost:42002/json/monitors.json',
	hostAccount: 'HOST_ACCOUNT',
};